/*! 
  @file rx_cu_common.cuh
	
  @brief CUDAʃwb_
 
  @author Makoto Fujisawa
  @date 2009-08, 2011-06
*/
// FILE --rx_cu_common.cuh--

#ifndef _RX_CU_COMMON_CUH_
#define _RX_CU_COMMON_CUH_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
//#include "cutil_math.h"
#include "vector_types.h"
#include "vector_functions.h"


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
typedef unsigned int uint;
typedef unsigned char uchar;

#define FLOAT float
#define FLOAT3 float3

#define MAKE_FLOAT3 make_float3

#define RX_CUMC_USE_GEOMETRY

#define RX_USE_ATOMIC_FUNC // vCompute capability 1.1ȏ(-arch sm_11)


// eNX`̎gptO
#ifndef __DEVICE_EMULATION__
#define USE_TEX 0
#endif

#if USE_TEX
#define FETCH(t, i) tex1Dfetch(t##Tex, i)
#else
#define FETCH(t, i) t[i]
#endif

#if USE_TEX
#define FETCHC(t, i) tex1Dfetch(t##Tex, i)
#else
#define FETCHC(t, i) cell.t[i]
#endif


// 1ubÑXbh(/)
#define BLOCK_SIZE 16

// 1ubN̍őXbh
#define THREAD_NUM 256

// Tv{[pꍇ1, A֐pꍇ0
#define SAMPLE_VOLUME 1

// Shared Memory̎gptO
#define USE_SHARED 1

// Shared Memorypꍇ̃Xbh̐
#define NTHREADS 32

#define SKIP_EMPTY_VOXELS 1

// CWT
#define M_PI 3.141592653589793238462643383279502884197
#define M_2PI 6.283185307179586476925286766559005768394		// 2*PI
#define M_SQRTPI 1.772453850905516027298167483341145182797	// sqrt(PI)
#define M_SQRT2PI 2.506628274631000502415765284811045253006	// sqrt(2*PI)

#define MEXICAN_HAT_FC (1.0/M_PI)
#define MEXICAN_HAT_C 0.8673250705840776f // c =  2/(sqrt(3)*pi^(1/4))
#define MEXICAN_HAT_R 5.0f

#define RX_MAX_FILTER_SIZE 10
#define RX_BINOMIALS_SIZE (RX_MAX_FILTER_SIZE+1)*(RX_MAX_FILTER_SIZE+1)


// s
struct matrix3x3
{
	float3 e[3];
};

struct matrix4x4
{
	float4 e[4];
};


#define MAX_BOX_NUM 10
#define MAX_SPHERE_NUM 10


//-----------------------------------------------------------------------------
//! SPHV~[Vp[^
//-----------------------------------------------------------------------------
struct rxSimParams
{
    FLOAT3 Gravity;
    FLOAT GlobalDamping;
    FLOAT ParticleRadius;

	float3 BoundaryMax;
	float3 BoundaryMin;

    uint3 GridSize;
    uint NumCells;
    FLOAT3 WorldOrigin;
	FLOAT3 WorldMax;
    FLOAT3 CellWidth;

    uint NumBodies;
    uint MaxParticlesPerCell;

	FLOAT EffectiveRadius;
	FLOAT Mass;			// p[eBN[kg]
	FLOAT VorticityConfinement;

	FLOAT Buoyancy;

	FLOAT Density;		// x[kg/m^3]
    FLOAT Pressure;     // [Pa = N/m^2 = kg/m.s^2]

    FLOAT Tension;		// [N/m = kg/s^2]
    FLOAT Viscosity;	// [Pa.s = N.s/m^2 = kg/m.s]
    FLOAT GasStiffness;	// [J = N.m = kg.m^2/s^2]  // used for DC96 symmetric pressure force

	FLOAT Volume;
	FLOAT KernelParticles;
	FLOAT Restitution;

	FLOAT Threshold;

	FLOAT InitDensity;

	FLOAT Dt;

	FLOAT Wpoly6;		//!< Pory6J[l̒萔W
	FLOAT GWpoly6;		//!< Pory6J[ľz̒萔W
	FLOAT LWpoly6;		//!< Pory6J[l̃vVA̒萔W
	FLOAT Wspiky;		//!< SpikyJ[l̒萔W
	FLOAT GWspiky;		//!< SpikyJ[ľz̒萔W
	FLOAT LWspiky;		//!< SpikyJ[l̃vVA̒萔W
	FLOAT Wvisc;		//!< ViscosityJ[l̒萔W
	FLOAT GWvisc;		//!< ViscosityJ[ľz̒萔W
	FLOAT LWvisc;		//!< ViscosityJ[l̃vVA̒萔W

	FLOAT Wd2;		//!< 
	FLOAT Wd3;		//!< 
	FLOAT GWd2;		//!< 
	FLOAT GWd3;		//!< 
	FLOAT Wd1;		//!< 

	uint   BoxNum;
#if MAX_BOX_NUM
	FLOAT3 BoxCen[MAX_BOX_NUM];
	FLOAT3 BoxExt[MAX_BOX_NUM];
	matrix3x3 BoxRot[MAX_BOX_NUM];
	matrix3x3 BoxInvRot[MAX_BOX_NUM];
	uint   BoxFlg[MAX_BOX_NUM];
#endif

	uint SphereNum;
#if MAX_SPHERE_NUM
	FLOAT3 SphereCen[MAX_SPHERE_NUM];
	FLOAT  SphereRad[MAX_SPHERE_NUM];
	uint   SphereFlg[MAX_SPHERE_NUM];
#endif
};

struct rxParticleCell
{
	float4* dSortedPos;			//!< \[gς݃p[eBNW
	float4* dSortedVel;			//!< \[gς݃p[eBNx

	uint* dSortedIndex;			//!< \[gς݃p[eBNCfbNX
	uint* dGridParticleHash;	//!< ep[eBÑObhnbVl(\[gpL[)
	uint* dCellStart;			//!< \[gXg̊eZ̃X^[gCfbNX
	uint* dCellEnd;				//!< \[gXg̊eZ̃GhCfbNX
	uint  uNumParticles;		//!< p[eBN
	uint  uNumCells;			//!< Z
	uint  uNumArdGrid;			//!< ߖTTQƃObh͈

	uint* dSortedPolyIdx;		//!< \[gς݃|SCfbNX
	uint* dGridPolyHash;		//!< |S̃ObhnbVl(\[gpL[)
	uint* dPolyCellStart;		//!< \[gXg̊eZ̃X^[gCfbNX
	uint* dPolyCellEnd;			//!< \[gXg̊eZ̃GhCfbNX
	uint  uNumPolyHash;
};




//-----------------------------------------------------------------------------
//! pbN邽߂̃f[^\ 
//-----------------------------------------------------------------------------
template<class T> 
struct rxVPack
{
	T* dPos;
	T* dCompactedPos;
	uint* dOcc;
	uint* dOccScan;

	rxVPack()
	{
		dPos = 0;
		dCompactedPos = 0;
		dOcc = 0;
		dOccScan = 0;
	}
};

typedef rxVPack<float> rxVPackf;

struct rxVrtAdd
{
	int num;
	int layer;
	int edge[2];
	float3 vrts[2];
};



#endif // _RX_CU_COMMON_CUH_
