/*! 
  @file rx_cu_funcs.cu
	
  @brief CUDA֐ - ֌WȂ

  @author Makoto Fujisawa
  @date 2009-08, 2011-06
*/
// FILE --rx_cu_funcs.cu--


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cutil_inline.h>
#include <cstdio>

#include <GL/glew.h>

#include <cutil.h>
#include <cuda_gl_interop.h>

#include <GL/glut.h>

#include "rx_cu_common.cu"


//-----------------------------------------------------------------------------
// CUDA֐
//-----------------------------------------------------------------------------
extern "C"
{
/*!
 * CUDAfoCX̐ݒ
 *  - R}hCɊÂCUDAfoCXݒ(()-device 0)
 * @param[in] argc R}hC̐
 * @param[in] argv R}hCXg(argv[0]͎st@C)
 */
void CuInit(int argc, char **argv)
{   
	if(cutCheckCmdLineFlag(argc, (const char**)argv, "device")){
		cutilDeviceInit(argc, argv);
	}
	else{
		cudaSetDevice( cutGetMaxGflopsDeviceId() );
	}
}

/*!
 * CUDAfoCX̐ݒ - id𒼐ڎw
 * @param[in] id foCXID
 */
void CuSetDevice(int id)
{ 
	int device_count = 0;
	cudaGetDeviceCount(&device_count);
	if(id < 0 || id >= device_count){
		id = cutGetMaxGflopsDeviceId();
		cudaSetDevice(id);
	}
	else{
		cudaSetDevice(id);
	}
}

/*!
 * foCX̊m
 * @param[out] dPtr foCXւ̃|C^
 * @param[in] size mۃTCY(̃TCY)
 */
void CuAllocateArray(void **dPtr, size_t size)
{
	cutilSafeCall(cudaMalloc(dPtr, size));
}

/*!
 * foCX̉
 * @param[in] devPtr foCXւ̃|C^
 */
void CuFreeArray(void *dPtr)
{
	cutilSafeCall(cudaFree(dPtr));
}

/*!
 * foCẌ̏
 * @param[in] dPtr foCXւ̃|C^
 * @param[in] val l
 * @param[in] size ̈̃TCY(̃TCY)
 */
void CuSetArrayValue(void *dPtr, int val, size_t size)
{
	cutilSafeCall(cudaMemset(dPtr, val, size));
}

/*!
 * foCXԃRs[
 * @param[in] dDst Rs[
 * @param[in] dSrc Rs[
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayD2D(void *dDst, void *dSrc, int size)
{
	cutilSafeCall(cudaMemcpy(dDst, dSrc, size, cudaMemcpyDeviceToDevice));
}

/*!
 * foCXzXgւ̃Rs[
 * @param[in] hDst Rs[zXg(ŒsizemۂĂ邱)
 * @param[in] dSrc Rs[foCX
 * @param[in] vbo dSrcVBȌꍇCVBOIDDłȂꍇ0w
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayFromDevice(void* hDst, const void* dSrc, unsigned int vbo, int size)
{   
	if(vbo) cutilSafeCall(cudaGLMapBufferObject((void**)&dSrc, vbo));

	cutilSafeCall(cudaMemcpy(hDst, dSrc, size, cudaMemcpyDeviceToHost));
	
	if(vbo) cutilSafeCall(cudaGLUnmapBufferObject(vbo));
}

/*!
 * zXgfoCXւ̃Rs[
 * @param[in] dDst Rs[foCX(ŒsizemۂĂ邱)
 * @param[in] hSrc Rs[zXg
 * @param[in] offset Rs[ItZbg
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayToDevice(void* dDst, const void* hSrc, int offset, int size)
{
	cutilSafeCall(cudaMemcpy((char*)dDst+offset, hSrc, size, cudaMemcpyHostToDevice));
}

/*!
 * VBO}bsO
 * @param[in] vbo VBO,PBO
 */
void *CuMapGLBufferObject(uint vbo)
{
	void *ptr;
	cutilSafeCall(cudaGLMapBufferObject(&ptr, vbo));
	return ptr;
}

/*!
 * VBOA}bv
 * @param[in] vbo VBO,PBO
 */
void CuUnmapGLBufferObject(uint vbo)
{
	cutilSafeCall(cudaGLUnmapBufferObject(vbo));
}

/*!
 * PBO,VBOobt@CUDAɓo^
 * @param[in] vbo VBO,PBO
 */
void CuRegisterGLBufferObject(uint vbo)
{
	cutilSafeCall(cudaGLRegisterBufferObject(vbo));
}

/*!
 * PBO,VBOobt@CUDA폜
 * @param[in] vbo VBO,PBO
 */
void CuUnregisterGLBufferObject(uint vbo)
{
	cutilSafeCall(cudaGLUnregisterBufferObject(vbo));
}

/*!
 * Xbh
 */
void CuThreadSync(void)
{
	cutilSafeCall(cudaThreadSynchronize());
}

/*!
 * foCXvpeB̕\
 */
void CuDeviceProp(void)
{
	int n;	//foCX
	cutilSafeCall(cudaGetDeviceCount(&n));

	for(int i = 0; i < n; ++i){
		cudaDeviceProp dev;

		// foCXvpeB擾
		cutilSafeCall(cudaGetDeviceProperties(&dev, i));

		printf("device %d\n", i);
		printf(" device name : %s\n", dev.name);
		printf(" total global memory : %d (MB)\n", dev.totalGlobalMem/1024/1024);
		printf(" shared memory / block : %d (KB)\n", dev.sharedMemPerBlock/1024);
		printf(" register / block : %d\n", dev.regsPerBlock);
		printf(" warp size : %d\n", dev.warpSize);
		printf(" max pitch : %d (B)\n", dev.memPitch);
		printf(" max threads / block : %d\n", dev.maxThreadsPerBlock);
		printf(" max size of each dim. of block : (%d, %d, %d)\n", dev.maxThreadsDim[0], dev.maxThreadsDim[1], dev.maxThreadsDim[2]);
		printf(" max size of each dim. of grid  : (%d, %d, %d)\n", dev.maxGridSize[0], dev.maxGridSize[1], dev.maxGridSize[2]);
		printf(" clock rate : %d (MHz)\n", dev.clockRate/1000);
		printf(" total constant memory : %d (KB)\n", dev.totalConstMem/1024);
		printf(" compute capability : %d.%d\n", dev.major, dev.minor);
		printf(" alignment requirement for texture : %d\n", dev.textureAlignment);
		printf(" device overlap : %s\n", (dev.deviceOverlap ? "ok" : "not"));
		printf(" num. of multiprocessors : %d\n", dev.multiProcessorCount);
		printf(" kernel execution timeout : %s\n", (dev.kernelExecTimeoutEnabled ? "on" : "off"));
		printf(" integrated : %s\n", (dev.integrated ? "on" : "off"));
		printf(" host memory mapping : %s\n", (dev.canMapHostMemory ? "on" : "off"));

		printf(" compute mode : ");
		if(dev.computeMode == cudaComputeModeDefault) printf("default mode (multiple threads can use) \n");
		else if(dev.computeMode == cudaComputeModeExclusive) printf("exclusive mode (only one thread will be able to use)\n");
		else if(dev.computeMode == cudaComputeModeProhibited) printf("prohibited mode (no threads can use)\n");
		
	}

	printf("Device with Maximum GFLOPS : %d\n", cutGetMaxGflopsDeviceId());
}


}   // extern "C"
