/*!
  @file rx_fltk_window.cpp
	
  @brief FLTKɂEBhENX
 
  @author Makoto Fujisawa 
  @date   2011-08
*/


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_fltk_window.h"
#include "rx_atom_ini.h"

#include "rx_sph.h"

//-----------------------------------------------------------------------------
// ϐ
//-----------------------------------------------------------------------------
// ݒt@C
rxINI *g_pINI = new rxINI;

// ftHgǂݍ݃t@C
vector<string> g_vDefaultFiles;

// Anisotropic Kernel̃p[^
double g_fLambda = 0.9;		//!< p[eBNʒuW[0,1]
double g_fRadScaling = 2.0;	//!< ߖTTa̔{
double g_fKr = 4.0;			//!< ŗLl`FbNpW(>1)
int g_iNe = 10;				//!< Ǘp[eBNp̋ߖTp[eBN
double g_fKn = 0.5;			//!< Ǘp[eBNɑ΂XP[OW
double g_fKs = 1400.0;		//!< p[eBN̑̐ς𒲐邽߂̌W


//-----------------------------------------------------------------------------
// rxFlWindowNX̎
//-----------------------------------------------------------------------------

//! RXgN^
rxFlWindow::rxFlWindow(int w_, int h_, const char* title)
	: Fl_Double_Window(w_, h_, title), m_iWinX(100), m_iWinY(100), m_iWinW(w_), m_iWinH(h_)
{
	m_pStatusLabel = 0;
	m_bFullScreen = false;

	resizable(this);

	int hs_menu = 20;	// j[o[̍
	int hs_para = 150;	// p[^pEBWbgzü̍
	int hs_stat = 24;	// Xe[^Xo[̍

	int hor_margin = 5;	// }[W
	int ver_margin = 5;	// }[W

	int xs = hor_margin;

	begin();
	{
		// `̈
		int ys = hs_menu+ver_margin;
		int ws = w()-hor_margin*2;
		int hs = h()-(hs_menu+hs_para+hs_stat+4*ver_margin);
		Fl_Group *g = new Fl_Group(xs, ys, ws, hs, 0);
		g->box(FL_NO_BOX);
		{
			m_pGLCanvas = new rxFlGLWindow(xs+2, ys+2, ws-4, hs-4, 0, this);
			m_pGLCanvas->box(FL_FLAT_BOX);
			m_pGLCanvas->mode(FL_RGB | FL_ALPHA | FL_DOUBLE | FL_DEPTH | FL_MULTISAMPLE);
			m_pGLCanvas->align(FL_ALIGN_INSIDE | FL_ALIGN_CENTER);
			cout << ws << ", " << hs << endl;

			// D&D{bNX
			m_pDndBox = new rxFlDndBox(xs, ys, ws, hs, 0);
			m_pDndBox->callback(OnDnd_s, this);
		}
		g->end();
		Fl_Group::current()->resizable(g);
	}

	ReadConfig(RX_PROGRAM_NAME);

	{
		// j[o[
		m_pMenuBar = new Fl_Menu_Bar(0, 0, w(), hs_menu, 0);

		// Filej[
		m_pMenuBar->add("File/Open", "^f", OnMenuFile_s, this); 
		m_pMenuBar->add("File/Save", FL_CTRL+FL_SHIFT+'s', OnMenuFile_s, this); 
		m_pMenuBar->add("File/Save FrameBuffer", FL_CTRL+'s', OnMenuFile_s, this); 
		m_pMenuBar->add("File/----------------"); 
		m_pMenuBar->add("File/Quit", FL_CTRL+'q', OnMenuFile_s, this); 

		// Drawj[
		int count = 0;
		while(RX_DRAW_STR[2*count] != "-1"){
			string label = "Draw/"+RX_DRAW_STR[2*count]+"  ";
			string shortcut = RX_DRAW_STR[2*count+1];
			m_pMenuBar->add(RX_TO_CHAR(label), RX_TO_CHAR(shortcut), rxFlGLWindow::OnMenuDraw_s, m_pGLCanvas, FL_MENU_TOGGLE);
			count++;
		}

		// Simulationj[
		m_pMenuBar->add("Simulation/Reset",						'R', rxFlGLWindow::OnMenuSimulation_s, m_pGLCanvas, FL_MENU_DIVIDER);
		m_pMenuBar->add("Simulation/Vorticity Confinement  ",	'V', rxFlGLWindow::OnMenuSimulation_s, m_pGLCanvas, FL_MENU_TOGGLE | FL_MENU_DIVIDER);
		m_pMenuBar->add("Simulation/Particle Data Input",		"^i", rxFlGLWindow::OnMenuSimulation_s, m_pGLCanvas, FL_MENU_TOGGLE);
		m_pMenuBar->add("Simulation/Particle Data Output",		"^o", rxFlGLWindow::OnMenuSimulation_s, m_pGLCanvas, FL_MENU_TOGGLE);
		m_pMenuBar->add("Simulation/Mesh Saving",				"^m", rxFlGLWindow::OnMenuSimulation_s, m_pGLCanvas, FL_MENU_TOGGLE);
		m_pMenuBar->add("Simulation/Image Saving",				"^a", rxFlGLWindow::OnMenuSimulation_s, m_pGLCanvas, FL_MENU_TOGGLE);

		// Particlej[
		m_pMenuBar->add("Particle/Anisotoropic Kernel  ",	'a', rxFlGLWindow::OnMenuParticle_s, m_pGLCanvas, FL_MENU_TOGGLE);

		// Particle/Colorj[
		m_pMenuBar->add("Particle/Color/Ramp",				0, rxFlGLWindow::OnMenuParticle_s, m_pGLCanvas, FL_MENU_RADIO);
		m_pMenuBar->add("Particle/Color/Density",			0, rxFlGLWindow::OnMenuParticle_s, m_pGLCanvas, FL_MENU_RADIO);
		m_pMenuBar->add("Particle/Color/Energy Spectrum",	0, rxFlGLWindow::OnMenuParticle_s, m_pGLCanvas, FL_MENU_RADIO);
		m_pMenuBar->add("Particle/Color/Pressure",			0, rxFlGLWindow::OnMenuParticle_s, m_pGLCanvas, FL_MENU_RADIO);
		m_pMenuBar->add("Particle/Color/None",				0, rxFlGLWindow::OnMenuParticle_s, m_pGLCanvas, FL_MENU_RADIO);

		// Particle/Drawj[
		count = 0;
		while(RX_PARTICLE_DRAW[2*count] != "-1"){
			string label = "Particle/Draw/"+RX_PARTICLE_DRAW[2*count]+"  ";
			string shortcut = RX_PARTICLE_DRAW[2*count+1];
			m_pMenuBar->add(RX_TO_CHAR(label), RX_TO_CHAR(shortcut), rxFlGLWindow::OnMenuParticle_s, m_pGLCanvas, FL_MENU_RADIO);
			count++;
		}

		// Solidj[
		count = 0;
		while(RX_SOLID_STR[2*count] != "-1"){
			string label = "Solid/"+RX_SOLID_STR[2*count]+"  ";
			string shortcut = RX_SOLID_STR[2*count+1];
			m_pMenuBar->add(RX_TO_CHAR(label), RX_TO_CHAR(shortcut), rxFlGLWindow::OnMenuSolid_s, m_pGLCanvas, FL_MENU_TOGGLE);
			count++;
		}

		// Meshj[
		count = 0;
		while(RX_TRIANGULATION_METHOD[2*count] != "-1"){
			string label = "Mesh/"+RX_TRIANGULATION_METHOD[2*count]+"  ";
			string shortcut = RX_TRIANGULATION_METHOD[2*count+1];
			m_pMenuBar->add(RX_TO_CHAR(label), RX_TO_CHAR(shortcut), rxFlGLWindow::OnMenuTriangulation_s, m_pGLCanvas, FL_MENU_RADIO);
			count++;
		}

		// Scenej[
		count = 0;
		while(!m_pGLCanvas->m_vSceneTitles[count].empty() && count < 12){
			string label = "Scene/"+m_pGLCanvas->m_vSceneTitles[count]+"  ";
			m_pMenuBar->add(RX_TO_CHAR(label), FL_F+count+1, rxFlGLWindow::OnMenuScene_s, m_pGLCanvas, FL_MENU_RADIO);
			count++;
		}
	
		// Stepj[
		m_pMenuBar->add("Step/Step", ' ', OnMenuStep_s, this); 
		m_pMenuBar->add("Step/Animation  ", 's', OnMenuStep_s, this); 

		// Windowj[
		m_pMenuBar->add("Window/FullScreen  ", FL_ALT+FL_Enter, OnMenuWindow_s, this, FL_MENU_TOGGLE); 

		// Helpj[
		m_pMenuBar->add("Help/Version  ", 0, OnMenuHelpVersion_s, this); 
	}
	{
		// GUÏ
		xs = hor_margin;
		int ys = h()-(hs_stat+ver_margin+hs_para);
		Fl_Scroll *g = new Fl_Scroll(hor_margin, ys, w()-2*hor_margin, hs_para, 0);
		g->box(FL_FLAT_BOX);

		xs += 7;
		ys += 5;
		int ws = 80;
		int hs = 25;

		Fl_Boxtype boxtype = FL_FLAT_BOX;
		Fl_Button *button;

		// Start/Stop{^
		button = new Fl_Button(xs, ys, ws, hs, "Start/Stop");
		button->callback(OnButtonStart_s, this);
		button->down_box(boxtype);
		button->clear_visible_focus();

		// Apply{^
		button = new Fl_Button(xs, ys+1*(hs+5), ws, hs, "Apply");
		button->callback(OnButtonApply_s, this);
		button->down_box(boxtype);
		button->clear_visible_focus();

		// Clear{^
		button = new Fl_Button(xs, ys+2*(hs+5), ws, hs, "Clear");
		button->down_box(boxtype);
		button->clear_visible_focus();

		//
		// V~[Vp[^
		//
		xs += ws+7;
		ys = h()-(hs_stat+ver_margin+hs_para)+5;
		ws = 310;
		hs = hs_para-6;

		Fl_Group* sg = new Fl_Group(xs, ys, ws, hs, "Anisotoropic");
		{
			sg->box(FL_DOWN_BOX);
			sg->labelsize(12);
			sg->align(FL_ALIGN_TOP | FL_ALIGN_INSIDE);

			int dx = 70;
			ys += 30;

			m_pSpinLambda = new Fl_Spinner(xs+dx, ys, 87, 25, "Lambda ");
			m_pSpinLambda->type(1);
			m_pSpinLambda->callback(OnSpinSimulation_s, this);
			m_pSpinLambda->minimum(0);
			m_pSpinLambda->maximum(1);
			m_pSpinLambda->step(0.01);
			m_pSpinLambda->value(g_fLambda);

			m_pSpinRadScaling = new Fl_Spinner(xs+dx, ys+30, 87, 25, "Rad ");
			m_pSpinRadScaling->type(1);
			m_pSpinRadScaling->callback(OnSpinSimulation_s, this);
			m_pSpinRadScaling->minimum(1);
			m_pSpinRadScaling->maximum(3);
			m_pSpinRadScaling->step(0.2);
			m_pSpinRadScaling->value(g_fRadScaling);
		
			m_pSpinKr = new Fl_Spinner(xs+dx, ys+60, 87, 25, "Kr ");
			m_pSpinKr->type(1);
			m_pSpinKr->callback(OnSpinSimulation_s, this);
			m_pSpinKr->minimum(0);
			m_pSpinKr->maximum(10);
			m_pSpinKr->step(0.1);
			m_pSpinKr->value(g_fKr);

			dx = 200;

			m_pSpinNe = new Fl_Spinner(xs+dx, ys, 87, 25, "Ne ");
			m_pSpinNe->type(1);
			m_pSpinNe->callback(OnSpinSimulation_s, this);
			m_pSpinNe->minimum(1);
			m_pSpinNe->maximum(50);
			m_pSpinNe->step(1);
			m_pSpinNe->value(g_iNe);
		
			m_pSpinKn = new Fl_Spinner(xs+dx, ys+30, 87, 25, "Kn ");
			m_pSpinKn->type(1);
			m_pSpinKn->callback(OnSpinSimulation_s, this);
			m_pSpinKn->minimum(0);
			m_pSpinKn->maximum(2);
			m_pSpinKn->step(0.02);
			m_pSpinKn->value(g_fKn);
		
			m_pSpinKs = new Fl_Spinner(xs+dx, ys+60, 87, 25, "Ks ");
			m_pSpinKs->type(1);
			m_pSpinKs->callback(OnSpinSimulation_s, this);
			m_pSpinKs->minimum(0);
			m_pSpinKs->maximum(2000);
			m_pSpinKs->step(10);
			m_pSpinKs->value(g_fKs);

			sg->resizable(NULL);
			sg->end();
		}

		xs += ws+5;
		ys = h()-(hs_stat+ver_margin+hs_para)+5;
		ws = 280;

		sg = new Fl_Group(xs, ys, ws, hs, "Draw");
		{
			sg->box(FL_DOWN_BOX);
			sg->labelsize(12);
			sg->align(FL_ALIGN_TOP | FL_ALIGN_INSIDE);

			int dx = 125;
			ys += 30;

			m_pSliderVScale = new Fl_Value_Slider(xs+dx, ys, 145, 25, "Veloc. Scale ");
			m_pSliderVScale->type(1);
			m_pSliderVScale->callback(OnSliderDraw_s, this);
			m_pSliderVScale->minimum(0.0);
			m_pSliderVScale->maximum(0.1);
			m_pSliderVScale->step(0.001);
			m_pSliderVScale->value(m_pGLCanvas->m_fVScale);
			m_pSliderVScale->align(Fl_Align(FL_ALIGN_LEFT));
			
			m_pSliderMeshThr = new Fl_Value_Slider(xs+dx, ys+30, 145, 25, "Mesh Threshold ");
			m_pSliderMeshThr->type(1);
			m_pSliderMeshThr->callback(OnSliderDraw_s, this);
			m_pSliderMeshThr->minimum(0.0);
			m_pSliderMeshThr->maximum(1000.0);
			m_pSliderMeshThr->step(5.0);
			m_pSliderMeshThr->value(m_pGLCanvas->m_fMeshThr);
			m_pSliderMeshThr->align(Fl_Align(FL_ALIGN_LEFT));

			m_pCheckMesh = new Fl_Check_Button(xs+dx, ys+60, 25, 25, "Surface Mesh ");
			m_pCheckMesh->down_box(FL_DOWN_BOX);
			m_pCheckMesh->callback(OnCheckDraw_s, this);
			m_pCheckMesh->align(Fl_Align(FL_ALIGN_LEFT));

			m_pCheckRefraction = new Fl_Check_Button(xs+dx, ys+80, 25, 25, "Refraction ");
			m_pCheckRefraction->down_box(FL_DOWN_BOX);
			m_pCheckRefraction->callback(OnCheckDraw_s, this);
			m_pCheckRefraction->align(Fl_Align(FL_ALIGN_LEFT));

			sg->resizable(NULL);
			sg->end();
		}

		g->resizable(NULL);
		g->end();
	}
	{
		// Xe[^Xo[(Fl_Box)
		int ys = h()-hs_stat;
		m_pBoxStatus = new Fl_Box(0, ys, w(), hs_stat, "status");
		m_pBoxStatus->box(FL_EMBOSSED_BOX);
		m_pBoxStatus->align(FL_ALIGN_INSIDE | FL_ALIGN_WRAP | FL_ALIGN_RIGHT);
		//m_pBoxStatus->color(color());
		m_pBoxStatus->labelfont(FL_HELVETICA_BOLD);
	}
	end();

	resize(m_iWinX, m_iWinY, m_iWinW, m_iWinH);

	UpdateMenuState();

	show();
}

//! fXgN^
rxFlWindow::~rxFlWindow()
{
	WriteConfig(RX_PROGRAM_NAME);

	if(m_pStatusLabel) delete [] m_pStatusLabel;

	if(m_pMenuBar) delete m_pMenuBar;

	if(m_pSpinLambda) delete m_pSpinLambda;
	if(m_pSpinRadScaling) delete m_pSpinRadScaling;
	if(m_pSpinKr) delete m_pSpinKr;
	if(m_pSpinNe) delete m_pSpinNe;
	if(m_pSpinKn) delete m_pSpinKn;
	if(m_pSpinKs) delete m_pSpinKs;

	if(m_pSliderMeshThr) delete m_pSliderMeshThr;
	if(m_pSliderVScale) delete m_pSliderVScale;

	if(m_pCheckMesh) delete m_pCheckMesh;
	if(m_pCheckRefraction) delete m_pCheckRefraction;

	if(m_pGLCanvas) delete m_pGLCanvas;
	if(m_pDndBox) delete m_pDndBox;
	if(m_pBoxStatus) delete m_pBoxStatus;
}


/*!
 * gOtj[ڂ̍XV
 */
void rxFlWindow::UpdateMenuState(void)
{
	int current, count;
	// Draw - FL_MENU_TOGGLE
	current = m_pGLCanvas->m_iDraw;
	count = 0;
	while(RX_DRAW_STR[2*count] != "-1"){
		string label = "Draw/"+RX_DRAW_STR[2*count]+"  ";
		SetMenuItemState(m_pMenuBar, RX_TO_CHAR(label), ((0x01 << count) & current), 1);
		count++;
	}

	bool turb = m_pGLCanvas->m_bsSimuSetting.at(ID_SPH_PS_TURB);

	// Simulation - FL_MENU_TOGGLE
	SetMenuItemState(m_pMenuBar, "Simulation/Vorticity Confinement  ",	(m_pGLCanvas->m_bsSimuSetting.at(ID_SPH_VC)));
	SetMenuItemState(m_pMenuBar, "Simulation/Particle Data Input",		(m_pGLCanvas->m_bsSimuSetting.at(ID_SPH_INPUT)));
	SetMenuItemState(m_pMenuBar, "Simulation/Particle Data Output",		(m_pGLCanvas->m_bsSimuSetting.at(ID_SPH_OUTPUT)));
	SetMenuItemState(m_pMenuBar, "Simulation/Mesh Saving",				(m_pGLCanvas->m_bsSimuSetting.at(ID_SPH_MESH_OUTPUT)));
	SetMenuItemState(m_pMenuBar, "Simulation/Image Saving",				(m_pGLCanvas->m_iSaveImageSpacing != -1));

	// Particle - FL_MENU_TOGGLE
	SetMenuItemState(m_pMenuBar, "Particle/Anisotoropic Kernel  ",	(m_pGLCanvas->m_bsSimuSetting.at(ID_SPH_ANISOTROPIC)));

	// Particle/Color - FL_MENU_RADIO
	SetMenuItemState(m_pMenuBar, "Particle/Color/Ramp",				(m_pGLCanvas->m_iColorType == rxParticleSystemBase::RX_RAMP));
	SetMenuItemState(m_pMenuBar, "Particle/Color/Density",			(m_pGLCanvas->m_iColorType == rxParticleSystemBase::RX_DENSITY));
	SetMenuItemState(m_pMenuBar, "Particle/Color/Pressure",			(m_pGLCanvas->m_iColorType == rxParticleSystemBase::RX_PRESSURE));
	SetMenuItemState(m_pMenuBar, "Particle/Color/None",				(m_pGLCanvas->m_iColorType == rxParticleSystemBase::RX_NONE));

	// Particle/Color - FL_MENU_RADIO
	current = m_pGLCanvas->m_iDrawPS;
	count = 0;
	while(RX_PARTICLE_DRAW[2*count] != "-1"){
		string label = "Particle/Draw/"+RX_PARTICLE_DRAW[2*count]+"  ";
		SetMenuItemState(m_pMenuBar, RX_TO_CHAR(label), (count == current), 1);
		count++;
	}

	// Solid - FL_MENU_TOGGLE
	current = m_pGLCanvas->m_iSolidDraw;
	count = 0;
	while(RX_SOLID_STR[2*count] != "-1"){
		string label = "Solid/"+RX_SOLID_STR[2*count]+"  ";
		SetMenuItemState(m_pMenuBar, RX_TO_CHAR(label), ((0x01 << count) & current), 1);
		count++;
	}

	// Mesh - FL_MENU_RADIO
	current = m_pGLCanvas->m_iTriangulationMethod;
	count = 0;
	while(RX_TRIANGULATION_METHOD[2*count] != "-1"){
		string label = "Mesh/"+RX_TRIANGULATION_METHOD[2*count]+"  ";
		SetMenuItemState(m_pMenuBar, RX_TO_CHAR(label), (count == current), 1);
		count++;
	}
	
	// Scene - FL_MENU_RADIO
	current = m_pGLCanvas->m_iCurrentSceneIdx;
	count = 0;
	while(!m_pGLCanvas->m_vSceneTitles[count].empty() && count < 12){
		string label = "Scene/"+m_pGLCanvas->m_vSceneTitles[count]+"  ";
		SetMenuItemState(m_pMenuBar, RX_TO_CHAR(label), (count == current), 1);
		count++;
	}

	// Windowj[
	SetMenuItemState(m_pMenuBar, "Window/FullScreen  ", m_bFullScreen, 1);

	// Refraction`FbN{^
	m_pCheckRefraction->value((m_pGLCanvas->m_iDraw & RX_IDD_REFRAC ? 1 : 0));

	// Mesh`FbN{^
	m_pCheckMesh->value(m_pGLCanvas->m_bsSimuSetting[ID_SPH_MESH]);
}

/*!
 * Fl_ButtoñR[obN֐ - Start/Stop{^
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnButtonStart_s(Fl_Widget *widget, void* x)
{
	((rxFlWindow*)x)->OnButtonStart();
}
void rxFlWindow::OnButtonStart(void)
{
	m_pGLCanvas->SwitchIdle(-1);
}

/*!
 * Fl_ButtoñR[obN֐ - Apply{^
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnButtonApply_s(Fl_Widget *widget, void* x)
{
	((rxFlWindow*)x)->OnButtonApply(widget);
}
void rxFlWindow::OnButtonApply(Fl_Widget *widget)
{
	m_pGLCanvas->m_fMeshThr = m_pSliderMeshThr->value();
	m_pGLCanvas->m_fVScale = m_pSliderVScale->value();
}

/*!
 * Fl_Spinner̃R[obN֐ - Simulation
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnSpinSimulation_s(Fl_Widget *widget, void* x)
{
	((rxFlWindow*)x)->OnSpinSimulation(widget);
}
void rxFlWindow::OnSpinSimulation(Fl_Widget *widget)
{
	Fl_Spinner *spin = (Fl_Spinner*)widget;
	string label = spin->label();
	double val = spin->value();

	if(label.find("Lambda") != string::npos){
		g_fLambda = m_pSpinLambda->value();
	}
	else if(label.find("Rad") != string::npos){
		g_fRadScaling = m_pSpinRadScaling->value();
	}
	else if(label.find("Kr") != string::npos){
		g_fKr = m_pSpinKr->value();
	}
	else if(label.find("Ne") != string::npos){
		g_iNe = (int)(m_pSpinNe->value()+0.01);
	}
	else if(label.find("Kn") != string::npos){
		g_fKn = m_pSpinKn->value();
	}
	else if(label.find("Ks") != string::npos){
		g_fKs = m_pSpinKs->value();
	}

}

/*!
 * Fl_Value_Slider̃R[obN֐ - Draw
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnSliderDraw_s(Fl_Widget *widget, void* x)
{
	((rxFlWindow*)x)->OnSliderDraw(widget);
}
void rxFlWindow::OnSliderDraw(Fl_Widget *widget)
{
	Fl_Value_Slider *slider = (Fl_Value_Slider*)widget;
	string label = slider->label();
	double val = slider->value();

	if(label.find("Veloc. Scale") != string::npos){
		m_pGLCanvas->m_fVScale = val;
	}
	else if(label.find("Mesh Threshold") != string::npos){
		m_pGLCanvas->m_fMeshThr = val;
	}
}

/*!
 * Fl_Check_ButtoñR[obN֐ - Draw
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnCheckDraw_s(Fl_Widget *widget, void* x)
{
	((rxFlWindow*)x)->OnCheckDraw(widget);
}
void rxFlWindow::OnCheckDraw(Fl_Widget *widget)
{
	Fl_Check_Button *check = (Fl_Check_Button*)widget;
	string label = check->label();
	int val = (int)(check->value()+0.001);

	if(label.find("Refraction") != string::npos){
		m_pGLCanvas->m_iDraw ^= RX_IDD_REFRAC;
	}
	else if(label.find("Surface Mesh") != string::npos){
		m_pGLCanvas->SetMeshCreation((val ? true : false));
	}

	m_pGLCanvas->redraw();
	UpdateMenuState();
}


/*!
 * rxFlDndBox̃R[obN֐
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnDnd_s(Fl_Widget *widget, void* x)
{
	((rxFlWindow*)x)->OnDnd();
}
void rxFlWindow::OnDnd(void)
{
	if(m_pDndBox->Event() == FL_PASTE){
		int dnd_text_len = m_pDndBox->EventTextLength();
		string dnd_text = m_pDndBox->EventText();

		// eLXg\nŕ
		vector<string> fns;
		size_t pos0 = 0, pos1 = 0;
		do{
			pos1 = dnd_text.find('\n', pos0);
			string fn = dnd_text.substr(pos0, pos1-pos0);
			if(fn.empty()) break;

			fns.push_back(fn);
			pos0 = pos1+1;
		}while(pos1 != string::npos);

		int n = (int)fns.size();
		for(int i = 0; i < n; ++i){
			cout << "file" << i << " : " << fns[i] << endl;
		}

		if(n == 0) return;

		for(int i = 0; i < n; ++i){
			Open(fns[i]);
		}
	}
}

/*!
 * Filej[R[obN֐
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnMenuFile_s(Fl_Widget *widget, void* x)
{
	Fl_Menu_Bar *menubar = (Fl_Menu_Bar*)widget;
	char picked[80];
	menubar->item_pathname(picked, sizeof(picked)-1);	// j[

	string label = picked;
	string menu_name = "Draw/";
	label = label.substr(menu_name.size(), string::npos);

	if(label.find("Open") == 0){
		((rxFlWindow*)x)->OnMenuFileOpen();
	}
	else if(label.find("Save") == 0){
		((rxFlWindow*)x)->OnMenuFileSave();
	}
	else if(label.find("Save FrameBuffer") == 0){
		((rxFlWindow*)x)->OnMenuFileSaveFrame();
	}
	else if(label.find("Quit") == 0){
		((rxFlWindow*)x)->OnMenuFileQuit();
	}
}

void rxFlWindow::OnMenuFileOpen(void)
{
	string filter = "3D Files|*.obj;*.dxf;*.wrl;*.3ds;*.stl;*.ply|All Files|*.*";
	vector<string> fns;
	int n = ShowFileDialog(fns, "Open 3D file", filter, false);

	if(n > 0){
		for(int i = 0; i < n; ++i){
			Open(fns[i]);
		}
	}
}
void rxFlWindow::OnMenuFileSave(void)
{
	string filter = "3D Files|*.obj;*.dxf;*.wrl;*.3ds;*.stl;*.ply|All Files|*.*";
	vector<string> fns;
	int n = ShowFileDialog(fns, "Save 3D file", filter, false);

	if(n > 0){
		for(int i = 0; i < n; ++i){
			Save(fns[i]);
		}
	}
}
void rxFlWindow::OnMenuFileSaveFrame(void)
{
	string filter = "Image Files|*.bmp;*.png|All Files|*.*";
	vector<string> fns;
	int n = ShowFileDialog(fns, "Save FrameBuffer", filter, false);

	if(n > 0){
		for(int i = 0; i < n; ++i){
			string ext = GetExtension(fns[i]);
			if(ext.empty()){
				fns[i] += ".png";
			}

			Save(fns[i]);
		}
	}
}
void rxFlWindow::OnMenuFileQuit(void)
{
	WriteConfig(RX_PROGRAM_NAME);
	exit(0);
}


/*!
 * Drawj[̃R[obN֐
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnMenuStep_s(Fl_Widget *widget, void* x)
{
	Fl_Menu_Bar *menubar = (Fl_Menu_Bar*)widget;
	char picked[80];
	menubar->item_pathname(picked, sizeof(picked)-1);	// j[

	string label = picked;
	string menu_name = "Step/";
	label = label.substr(menu_name.size(), string::npos);
	
	((rxFlWindow*)x)->OnMenuStep(label);
}
void rxFlWindow::OnMenuStep(string label)
{
	if(label.find("Step") == 0){// Aj[V1Xebvs
		m_pGLCanvas->Idle();
	}
	else if(label.find("Animation") == 0){	// Aj[VON/OFF
		m_pGLCanvas->SwitchIdle(-1);
	}
}


/*!
 * Windowj[̃R[obN֐
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnMenuWindow_s(Fl_Widget *widget, void* x)
{
	Fl_Menu_Bar *menubar = (Fl_Menu_Bar*)widget;
	char picked[80];
	menubar->item_pathname(picked, sizeof(picked)-1);	// j[

	string label = picked;
	string menu_name = "Window/";
	label = label.substr(menu_name.size(), string::npos);
	
	((rxFlWindow*)x)->OnMenuWindow(label);
}
void rxFlWindow::OnMenuWindow(string label)
{
	if(label.find("FullScreen") == 0){	// tXN[ON/OFF
		SwitchFullScreen();
	}
}

/*!
 * j[:Help -> VersioñR[obN֐
 * @param[in] widget EBWbg̐eNXIuWFNg
 * @param[in] x [U`ϐ
 */
void rxFlWindow::OnMenuHelpVersion_s(Fl_Widget *widget, void* x)
{
	fl_message("OpenGL Application by FLTK\n  version 1.0");
}

/*!
 * Xe[^Xo[ɕݒ
 * @param[in] label \
 */
void rxFlWindow::SetStatusLabel(const string &label)
{
	if(m_pStatusLabel) delete [] m_pStatusLabel;
	m_pStatusLabel = RX_TO_CHAR(label);
	m_pBoxStatus->label(m_pStatusLabel);
}

/*!
 * t@Cǂݍ
 * @param[in] fn t@CpX
 */
void rxFlWindow::Open(const string &fn)
{
	string ext = GetExtension(fn);

	if(ext == "obj" || ext == "dxf" || ext == "wrl" || ext == "3ds" || ext == "stl" || ext == "ply"){
		m_pGLCanvas->OpenFile(fn);
	}
	else if(ext == "bmp" || ext == "jpg" || ext == "png" || ext == "gif" || ext == "tif"){
		return;
	}

	// ǂݍ񂾃t@Ci[
	m_strFullPath = fn;
	m_strFileName = GetFileName(fn);

	// t@CXe[^Xo[ɕ\
	SetStatusLabel(m_strFileName);
}

/*!
 * t@C
 * @param[in] fn t@CpX
 */
void rxFlWindow::Save(const string &fn)
{
	string ext = GetExtension(fn);

	if(ext == "obj" || ext == "dxf" || ext == "wrl" || ext == "3ds" || ext == "stl" || ext == "ply"){
		m_pGLCanvas->SaveFile(fn);

		// ǂݍ񂾃t@Ci[
		m_strFullPath = fn;
		m_strFileName = GetFileName(fn);

		// t@CXe[^Xo[ɕ\
		SetStatusLabel(m_strFileName);
	}
	else if(ext == "bmp" || ext == "png"){
		m_pGLCanvas->SaveDisplay(fn);
	}
}



/*!
 * ݒt@Cǂݍ
 * @param[in] fn ݒt@C(gq)
 */
void rxFlWindow::ReadConfig(const string &fn)
{
	// AvP[VǗݒt@C
	if(!g_pINI) g_pINI = new rxINI();

	g_pINI->Set("window", "width",  &m_iWinW, m_iWinW);
	g_pINI->Set("window", "height", &m_iWinH, m_iWinH);
	g_pINI->Set("window", "pos_x",  &m_iWinX, m_iWinX);
	g_pINI->Set("window", "pos_y",  &m_iWinY, m_iWinY);

	if(!(g_pINI->Load(fn+".ini"))){
		cout << "Failed opening the " << fn << ".ini file!" << endl;
	}
}
/*!
 * ݒt@C
 * @param[in] fn ݒt@C(gq)
 */
void rxFlWindow::WriteConfig(const string &fn)
{
	m_iWinW = w();
	m_iWinH = h();
	m_iWinX = x();
	m_iWinY = y();
	if(g_pINI->Save(fn+".ini")){
		cout << "save : " << fn << ".ini" << endl;
	}
}

/*!
 * tXN[/EBhE\̐؂ւ
 */
void rxFlWindow::SwitchFullScreen(void)
{
	static int pos0[2] = { 0, 0 };
	static int win0[2] = { 500, 500 };
	if(m_bFullScreen){
		fullscreen_off(pos0[0], pos0[1], win0[0], win0[1]);
		m_bFullScreen = false;
	}
	else{
		pos0[0] = x();
		pos0[1] = y();
		win0[0] = w();
		win0[1] = h();
		fullscreen();
		m_bFullScreen = true;
	}

}

/*!
 * Cxgnh
 * @param[in] ev CxgID
 */
int rxFlWindow::handle(int ev)
{
	switch(ev){
	case FL_DND_ENTER:
	case FL_DND_RELEASE:
	case FL_DND_LEAVE:
	case FL_DND_DRAG:
	case FL_PASTE:
		return 1;

	case FL_PUSH:		// }EX{^_E
		m_pGLCanvas->Mouse(Fl::event_button(), 1, Fl::event_x(), Fl::event_y());
		break;
	case FL_RELEASE:	// }EX{^Abv
		m_pGLCanvas->Mouse(Fl::event_button(), 0, Fl::event_x(), Fl::event_y());
		break;
	case FL_DRAG:		// }EXhbO
		m_pGLCanvas->Motion(Fl::event_x(), Fl::event_y());
		break;
	case FL_MOVE:		// }EXړ
		m_pGLCanvas->PassiveMotion(Fl::event_x(), Fl::event_y());
		break;

	case FL_KEYBOARD:	// L[_E
		m_pGLCanvas->Keyboard(Fl::event_key(), Fl::event_x(), Fl::event_y());
		UpdateMenuState();
		break;

	case FL_SHORTCUT:	// O[oV[gJbg
		break;

	default:
		break;
	}

	return Fl_Window::handle(ev);
}