/*!
  @file rx_ps.h
	
  @brief p[eBNV~[V̊NX
 
  @author Makoto Fujisawa
  @date 2011-06
*/
// FILE --rx_ps.h--

#ifndef _RX_PS_H_
#define _RX_PS_H_


//-----------------------------------------------------------------------------
// MARK:CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_sph_commons.h"

#include "rx_cu_common.cuh"
#include "radixsort.h"

#include "rx_sph_solid.h"

#include <cutil_inline.h>
//#include <cutil_math.h>



//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
const int DIM = 4;
const int RX_MAX_STEPS = 100000;


//-----------------------------------------------------------------------------
// p[eBNC
//-----------------------------------------------------------------------------
struct rxInletLine
{
	Vec3 pos1, pos2;	//!< C̒[_
	Vec3 vel;			//!< ǉp[eBN̑x
	Vec3 up;			//!< p[eBN͐ϕ
	int accum;			//!< p[eBN͐ϐ
	int span;			//!< ԓIȃXp
	double spacing;		//!< ԓIȃXp
};


//-----------------------------------------------------------------------------
// p[eBNV~[V̊NX
//-----------------------------------------------------------------------------
class rxParticleSystemBase
{
public:
	//! ߖTTʊi[
	struct rxNeigh
	{
		int Idx;		//!< ߖTp[eBNCfbNX
		RXREAL Dist;	//!< ߖTp[eBN܂ł̋
		RXREAL Dist2;	//!< ߖTp[eBN܂ł2拗
	};

	enum rxParticleConfig
	{
		RX_CONFIG_RANDOM,
		RX_CONFIG_GRID,
		RX_CONFIG_BOX,
		RX_CONFIG_NONE, 
		_NUM_CONFIGS
	};

	enum rxParticleArray
	{
		RX_POSITION = 0,
		RX_VELOCITY,
		RX_NORMAL, 
		RX_FORCE, 
		RX_DENSITY, 
		RX_PRESSURE, 

		RX_VORTICITY, 

		RX_UPDATED_POSITION, 
		RX_EIGEN_VALUE, 
		RX_ROTATION_MATRIX, 
		RX_TRANSFORMATION, 

		RX_RAMP, 
		RX_NONE, 

		RX_PSDATA_END, 
	};

protected:
	bool m_bInitialized;
	bool m_bUseOpenGL;

	uint m_uNumParticles;
	uint m_uMaxParticles;
	uint m_uNumArdGrid;

	uint m_solverIterations;

	RXREAL m_fParticleRadius;
	Vec3   m_v3Gravity;
	RXREAL m_fDamping;
	RXREAL m_fRestitution;

	RXREAL *m_hPos;		//!< p[eBNʒu
	RXREAL *m_hVel;		//!< p[eBNx

	uint m_posVBO;		//!< p[eBNWVBO
	uint m_colorVBO;	//!< J[VBO
	
	Vec3 m_v3EnvMin;	//!< AABBŏW
	Vec3 m_v3EnvMax;	//!< AABBőW

	int m_iColorType;

	RXREAL m_fTime;

	bool m_bUseVorticity;	//!< QxgptO
	bool m_bGridVelocity;	//!< Obhւ̑x꓊etO
	bool m_bCalNormal;		//!< @vZtO


protected:
	rxParticleSystemBase(){}

public:
	//! RXgN^
	rxParticleSystemBase(bool bUseOpenGL) : 
		m_bInitialized(false),
		m_bUseOpenGL(bUseOpenGL), 
		m_hPos(0),
		m_hVel(0)
	{
		m_v3Gravity = Vec3(0.0, -9.82, 0.0);
		m_fDamping = 0.0;
		m_fRestitution = 0.0;
		m_fParticleRadius = 0.1;
		m_fTime = 0.0;
		m_bCalAnisotropic = false;
	}

	//! fXgN^
	virtual ~rxParticleSystemBase(){}


	// V~[V
	Vec3 GetMax(void) const { return m_v3EnvMax; }
	Vec3 GetMin(void) const { return m_v3EnvMin; }
	Vec3 GetDim(void) const { return m_v3EnvMax-m_v3EnvMin; }
	Vec3 GetCen(void) const { return 0.5*(m_v3EnvMax+m_v3EnvMin); }

	// p[eBN
	int	GetNumParticles() const { return m_uNumParticles; }
	int	GetMaxParticles() const { return m_uMaxParticles; }

	// V~[V
	void SetIterations(int i) { m_solverIterations = i; }
		
	// p[eBNa
	float GetParticleRadius(){ return m_fParticleRadius; }

	// V~[Vݒ
	void SetDamping(RXREAL x){ m_fDamping = x; }	//!< ő̋Eł̔
	void SetGravity(RXREAL x){ m_v3Gravity = Vec3(0.0, x, 0.0); }	//!< d

	// p[eBNVBO
	unsigned int GetCurrentReadBuffer() const { return m_posVBO; }
	unsigned int GetColorBuffer()	   const { return m_colorVBO; }

	// `pJ[ݒ
	void SetColorType(int type){ m_iColorType = type; }
	int  GetColorType(void) const { return m_iColorType; }

	// tOؑ
	void ToggleUseVorticity(int t = -1){ RX_TOGGLE(m_bUseVorticity, t); }	//!< Qx
	bool IsUseVorticity(void) const { return m_bUseVorticity; }
	void ToggleGridVelocity(int t = -1){ RX_TOGGLE(m_bGridVelocity, t); }	//!< Obhւ̑x̓e
	bool IsGridVelocity(void) const { return m_bGridVelocity; }
	void ToggleNormalCalc(int t = -1){ RX_TOGGLE(m_bCalNormal, t); }		//!< p[eBN@̌vZ
	bool IsNormalCalc(void) const { return m_bCalNormal; }

public:
	// z֐
	virtual bool Update(RXREAL dt) = 0;

	virtual RXREAL* GetArrayVBO(rxParticleArray type, bool d2h = true) = 0;
	virtual void SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count) = 0;
	virtual void SetColorVBO(int type) = 0;

	virtual RXREAL* GetParticle(void) = 0;
	virtual RXREAL* GetParticleDevice(void) = 0;

protected:
	virtual void setObjectToCell(RXREAL *p) = 0;

public:
	// z֐
	virtual void UnmapParticle(void){}

	virtual void SetPolygonObstacle(const vector<Vec3> &vrts, const vector<Vec3> &nrms, const vector< vector<int> > &tris){}
	virtual void SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, int flg){}
	virtual void SetSphereObstacle(Vec3 cen, double rad, int flg){}
	virtual void MoveSphereObstacle(int b, Vec3 disp){}
	virtual Vec3 GetSphereObstaclePos(int b = -1){ return Vec3(0.0); }

	virtual void SetParticlesToCell(void) = 0;
	virtual void SetPolygonsToCell(void){}

	// A֐lvZ
	virtual double GetImplicit(double x, double y, double z){ return 0.0; }
	virtual void CalImplicitField(int n[3], Vec3 minp, Vec3 d, RXREAL *hF){}
	virtual void CalImplicitFieldDevice(int n[3], Vec3 minp, Vec3 d, RXREAL *dF){}

	// `֐
	virtual void DrawCell(int i, int j, int k){}
	virtual void DrawCells(Vec3 col, Vec3 col2, int sel = 0){}

	virtual void DrawObstacles(void){}

	// V~[Vݒ̏o
	virtual void OutputSetting(string fn){}

	// Anisotropic Kernel
	virtual void CalAnisotropicKernel(void){}
	bool m_bCalAnisotropic;

public:
	void Reset(rxParticleConfig config);
	bool Set(const vector<Vec3> &ppos, const vector<Vec3> &pvel);

	void AddSphere(int start, RXREAL *pos, RXREAL *vel, int r, RXREAL spacing);
	void AddBox(int start, Vec3 cen, Vec3 dim, Vec3 vel, RXREAL spacing);
	int  AddLine(int &start, rxInletLine line);

	RXREAL SetColorVBOFromArray(RXREAL *hVal, int d, bool use_max = true, RXREAL vmax = 1.0);
	void SetColorVBO(void){ SetColorVBO(m_iColorType); }

	int OutputParticles(string fn);
	int InputParticles(string fn);

protected:
	uint createVBO(uint size)
	{
		GLuint vbo;
		glGenBuffers(1, &vbo);
		glBindBuffer(GL_ARRAY_BUFFER, vbo);
		glBufferData(GL_ARRAY_BUFFER, size, 0, GL_DYNAMIC_DRAW);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		return vbo;
	}

};



#endif	// _PS_H_

