/*!
  @file rx_sph.h
	
  @brief SPH@
 
  @author Makoto Fujisawa
  @date 2008-10,2011-06
*/
// FILE --rx_sph.h--

#ifndef _RX_SPH_H_
#define _RX_SPH_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_sph_commons.h"

#include "rx_ps.h"

#include "rx_cu_common.cuh"
#include "radixsort.h"

#include "rx_sph_solid.h"

#include <cutil_inline.h>
//#include <cutil_math.h>



//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
//#define GL_REAL GL_DOUBLE
#define GL_REAL GL_FLOAT

// Ԍv
class rxTimerAvg;
extern rxTimerAvg g_Time;

#define RX_USE_TIMER

#ifdef RX_USE_TIMER
#define RXTIMER_CLEAR g_Time.ClearTime()
#define RXTIMER_RESET g_Time.ResetTime()
#define RXTIMER(x) g_Time.Split(x)
#define RXTIMER_PRINT g_Time.Print()
#define RXTIMER_STRING(x) g_Time.PrintToString(x)
#else
#define RXTIMER_CLEAR
#define RXTIMER_RESET
#define RXTIMER(x) 
#define RXTIMER_PRINT
#define RXTIMER_STRING(x)
#endif

// Anisotropic Kernel̃p[^
extern double g_fLambda;
extern double g_fRadScaling;
extern double g_fKr;
extern int g_iNe;
extern double g_fKn;
extern double g_fKs;



//! SPHV[̃p[^
struct rxSPHEnviroment
{
	int max_particles;			//!< őp[eBN
	Vec3 boundary_cen;			//!< E̒S
	Vec3 boundary_ext;			//!< Ȇ傫(eӂ̒1/2)
	RXREAL dens;				//!< x
	RXREAL mass;				//!< p[eBN̎
	RXREAL kernel_particles;	//!< Lahȓ̃p[eBN
};



//-----------------------------------------------------------------------------
// MARK:rxSPHNX̐錾
//-----------------------------------------------------------------------------
class rxSPH : public rxParticleSystemBase
{

	//! TpԕObh̊eZ
	struct rxCell
	{
		uint* hSortedIndex;			//!< nbVlŃ\[gp[eBNCfbNX
		uint* hGridParticleHash;	//!< ep[eBÑObhnbVl
		uint* hCellStart;			//!< \[gXg̊eZ̃X^[gCfbNX
		uint* hCellEnd;				//!< \[gXg̊eZ̃GhCfbNX
		uint  uNumCells;			//!< Z

		uint* hSortedPolyIdx;		//!< nbVlŃ\[g|SCfbNX(dL)
		uint* hGridPolyHash;		//!< e|S̃ObhnbVl
		uint* hPolyCellStart;		//!< \[gXg̊eZ̃X^[gCfbNX
		uint* hPolyCellEnd;			//!< \[gXg̊eZ̃GhCfbNX	

		uint uNumPolyHash;			//!< |S܂ރZ̐
	};

private:
	//
	// oϐ
	//

	// p[eBN
	RXREAL *m_hNrm;					//!< p[eBN@
	RXREAL *m_hFrc;					//!< p[eBNɂ
	RXREAL *m_hDens;				//!< p[eBNx
	RXREAL *m_hPres;				//!< p[eBN

	// \ʐp(Anisotropic kernel)
	RXREAL *m_hUpPos;				//!< p[eBNʒu
	RXREAL *m_hPosW;				//!< dݕtύW
	RXREAL *m_hCMatrix;				//!< Us
	RXREAL *m_hEigen;				//!< Usْ̓l
	RXREAL *m_hRMatrix;				//!< ]s(Us̓كxNg)
	RXREAL *m_hG;					//!< ό`s

	// EEő
	rxSolid *m_pBoundary;			//!< V~[VԂ̋E
	vector<rxSolid*> m_vSolids;		//!< ő̕
	RXREAL *m_hVrts;				//!< ő̃|S̒_
	int m_iNumVrts;					//!< ő̃|S̒_
	int *m_hTris;					//!< ő̃|S
	int m_iNumTris;					//!< ő̃|S̐

	// ԕiq֘A
	rxCell m_hCellData;				//!< TpԕObh
	int m_iGridSize[3];				//!< iq̐
	RXREAL m_fCellWidth[3];			//!< iqЂ̒
	vector< vector<rxNeigh> > m_vNeighs;	//!< ߖTp[eBN

	// qp[^
	uint m_iKernelParticles;		//!< J[l̃p[eBN
	RXREAL m_fInitDens, m_fMass;	//!< xC
	RXREAL m_fEffectiveRadius;		//!< La

	// V~[Vp[^
	RXREAL m_fGasStiffness;			//!< KX萔
	RXREAL m_fViscosity;			//!< SW
	RXREAL m_fBuoyancy;				//!< 

	// J[l֐̌vZ̍ۂɗp萔W
	double m_fWpoly6;				//!< Pory6J[l̒萔W
	double m_fGWpoly6;				//!< Pory6J[ľz̒萔W
	double m_fLWpoly6;				//!< Pory6J[l̃vVA̒萔W
	double m_fWspiky;				//!< SpikyJ[l̒萔W
	double m_fGWspiky;				//!< SpikyJ[ľz̒萔W
	double m_fLWspiky;				//!< SpikyJ[l̃vVA̒萔W
	double m_fWvisc;				//!< ViscosityJ[l̒萔W
	double m_fGWvisc;				//!< ViscosityJ[ľz̒萔W
	double m_fLWvisc;				//!< ViscosityJ[l̃vVA̒萔W

protected:
	rxSPH(){}

public:
	//! RXgN^
	rxSPH(bool use_opengl);

	//! fXgN^
	virtual ~rxSPH();

	// p[eBNa
	float GetEffectiveRadius(){ return m_fEffectiveRadius; }

	// ߖTp[eBN
	uint* GetNeighborList(const int &i, int &n);

public:
	//
	// z֐
	//
	// p[eBNf[^
	virtual RXREAL* GetParticle(void);
	virtual RXREAL* GetParticleDevice(void);
	virtual void UnmapParticle(void){}

	// V~[VXebv
	virtual bool Update(RXREAL dt);

	// V[̐ݒ
	virtual void SetPolygonObstacle(const vector<Vec3> &vrts, const vector<Vec3> &nrms, const vector< vector<int> > &tris);
	virtual void SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, int flg);
	virtual void SetSphereObstacle(Vec3 cen, double rad, int flg);
	virtual void MoveSphereObstacle(int b, Vec3 disp);
	virtual Vec3 GetSphereObstaclePos(int b = -1);

	// zXg<->VBOԓ]
	virtual RXREAL* GetArrayVBO(rxParticleArray type, bool d2h = true);
	virtual void SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count);
	virtual void SetColorVBO(int type);


	// A֐lvZ
	virtual double GetImplicit(double x, double y, double z);
	virtual void CalImplicitField(int n[3], Vec3 minp, Vec3 d, RXREAL *hF);
	virtual void CalImplicitFieldDevice(int n[3], Vec3 minp, Vec3 d, RXREAL *dF);

	// SPHo
	virtual void OutputSetting(string fn);

	// `֐
	virtual void DrawCell(int i, int j, int k);
	virtual void DrawCells(Vec3 col, Vec3 col2, int sel = 0);
	virtual void DrawObstacles(void);

protected:
	virtual void setObjectToCell(RXREAL *p);
	
	
public:
	// SPH
	void Initialize(const rxSPHEnviroment &env);
	void Allocate(int max_particles);
	void Finalize(void);

	// ߖT擾
	void GetNearestNeighborsDirect(int idx, RXREAL *prts, vector<rxNeigh> &neighs, RXREAL h = -1.0);
	void GetNearestNeighborsDirectP(Vec3 pos, RXREAL *prts, vector<rxNeigh> &neighs, RXREAL h = -1.0);
	void GetNearestNeighborsDirect(int idx, vector<rxNeigh> &neighs, RXREAL h = -1.0);

	void GetNearestNeighbors(int idx, vector<rxNeigh> &neighs, RXREAL h = -1.0);
	void GetNearestNeighbors(int idx, RXREAL *p, vector<rxNeigh> &neighs, RXREAL h = -1.0);
	void GetNearestNeighborsP(Vec3 pos, RXREAL *prts, vector<rxNeigh> &neighs, RXREAL h = -1.0);

	// ZɃp[eBNi[
	virtual void SetParticlesToCell(void);
	virtual void SetPolygonsToCell(void);

	int  GetPolygonsInCell(int gi, int gj, int gk, vector<int> &polys);
	bool IsPolygonsInCell(int gi, int gj, int gk);

	double CalColorField(double x, double y, double z);
	
	//
	// Anisotropic kernel
	//
	virtual void CalAnisotropicKernel(void);

protected:
	// CPUɂSPHvZ
	void calDensity(void);
	void calNormal(void);
	void calForce(void);

	// ʒuƑx̍XV(Leap-Frog)
	void integrate(const RXREAL *pos, const RXREAL *vel, const RXREAL *frc, 
				   RXREAL *pos_new, RXREAL *vel_new, RXREAL dt);

	// Փ˔
	int calCollisionPolygon(uint grid_hash, Vec3 &pos0, Vec3 &pos1, Vec3 &vel, RXREAL dt);
	int calCollisionSolid(Vec3 &pos0, Vec3 &pos1, Vec3 &vel, RXREAL dt);


	// Zփp[eBNi[
	void setObjectToCell(int pidx, RXREAL *p);
	
	// Z̏ݒ
	void setupCells(const Vec3 &vMin, const Vec3 &vMax, const double &h);

	// ZߖTp[eBN擾
	void getNeighborsInCell(Vec3 pos, RXREAL *prts, int gi, int gj, int gk, vector<rxNeigh> &neighs, RXREAL h);
	void searchNeighbors(RXREAL h = -1.0);
	void searchNeighbors(RXREAL *p, RXREAL h = -1.0);

	// ObhnbV̌vZ
	uint calGridHash(int x, int y, int z);
	uint calGridHash(Vec3 pos);


	// |S𕪊ZɊi[
	void setPolysToCell(RXREAL *vrts, int nv, int* tris, int nt);
};


//-----------------------------------------------------------------------------
// MARK:rxSPH_GPUNX̐錾
//-----------------------------------------------------------------------------
class rxSPH_GPU : public rxParticleSystemBase
{
private:
	//
	// oϐ(GPUϐ)
	//
	RXREAL *m_dPos;			//!< p[eBNʒu
	RXREAL *m_dVel;			//!< p[eBNx
	RXREAL *m_dNrm;			//!< p[eBN@
	RXREAL *m_dFrc;			//!< p[eBNɂ
	RXREAL *m_dDens;		//!< p[eBNx
	RXREAL *m_dPres;		//!< p[eBN

	// SPS
	RXREAL *m_dUpPos;		//!< p[eBNʒu
	RXREAL *m_dPosW;		//!< dݕtύW
	RXREAL *m_dCMatrix;		//!< Us
	RXREAL *m_dEigen;		//!< Usْ̓l
	RXREAL *m_dRMatrix;		//!< ]s(Us̓كxNg)
	RXREAL *m_dG;			//!< ό`s

	// \ʃbV
	RXREAL *m_dVrts;		//!< ő̃bV_
	int    *m_dTris;		//!< ő̃bV

	// V~[Vp[^
	rxSimParams m_params;	//!< V~[Vp[^(GPUւ̃f[^np)
	uint3 m_gridSize;		//!< ߖTTObh̊e̕
	uint m_numGridCells;	//!< ߖTTObh
	
	// ԕ(GPU)
	rxParticleCell m_dCellData;	//!< ߖTTObh
	uint m_gridSortBits;		//!< nbVlɂ\[g̊
	RadixSort *m_sorterForPrt;	//!< \[g(p[eBNp)
	RadixSort *m_sorterForPly;	//!< \[g(|Sp)

	//
	// oϐ(CPUϐ)
	//
	RXREAL *m_hNrm;			//!< p[eBN@
	RXREAL *m_hFrc;			//!< p[eBNɂ
	RXREAL *m_hDens;		//!< p[eBNx
	RXREAL *m_hPres;		//!< p[eBN

	// \ʐp(Anisotropic kernel)
	RXREAL *m_hUpPos;		//!< p[eBNʒu
	RXREAL *m_hPosW;		//!< dݕtύW
	RXREAL *m_hCMatrix;		//!< Us
	RXREAL *m_hEigen;		//!< Usْ̓l
	RXREAL *m_hRMatrix;		//!< ]s(Us̓كxNg)
	RXREAL *m_hG;			//!< ό`s
	RXREAL  m_fEigenMax;	//!< ْl̍ől(Tagɗp)

	// \ʃbV
	vector<RXREAL> m_vVrts;	//!< ő̃bV_
	int m_iNumVrts;			//!< ő̃bV_
	vector<int> m_vTris;	//!< ő̃bV
	int m_iNumTris;			//!< ő̃bV

	bool m_bCalNormal;		//!< @vZtO

protected:
	rxSPH_GPU(){}

public:
	//! RXgN^
	rxSPH_GPU(bool use_opengl);

	//! fXgN^
	~rxSPH_GPU();

	// p[eBNa
	float GetEffectiveRadius(){ return m_params.EffectiveRadius; }

	// ߖTp[eBN
	uint* GetNeighborList(const int &i, int &n);

	// V~[Vp[^
	rxSimParams GetParams(void){ return m_params; }
	void UpdateParams(void);

	// tOؑ
	void ToggleNormalCalc(int t = -1){ RX_TOGGLE(m_bCalNormal, t); }			//!< p[eBN@̌vZ
	bool IsNormalCalc(void) const { return m_bCalNormal; }
#if MAX_BOX_NUM
	void ToggleSolidFlg(int t = -1);
#endif

public:
	//
	// z֐
	//
	// p[eBNf[^
	virtual RXREAL* GetParticle(void);
	virtual RXREAL* GetParticleDevice(void);
	virtual void UnmapParticle(void);

	// V~[VXebv
	virtual bool Update(RXREAL dt);

	// V[̐ݒ
	virtual void SetPolygonObstacle(const vector<Vec3> &vrts, const vector<Vec3> &nrms, const vector< vector<int> > &tris);
	virtual void SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, int flg);
	virtual void SetSphereObstacle(Vec3 cen, double rad, int flg);
	virtual void MoveSphereObstacle(int b, Vec3 disp);
	virtual Vec3 GetSphereObstaclePos(int b = -1);

	// zXg<->VBOԓ]
	virtual RXREAL* GetArrayVBO(rxParticleArray type, bool d2h = true);
	virtual void SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count);
	virtual void SetColorVBO(int type);


	// A֐lvZ
	virtual double GetImplicit(double x, double y, double z);
	virtual void CalImplicitField(int n[3], Vec3 minp, Vec3 d, RXREAL *hF);
	virtual void CalImplicitFieldDevice(int n[3], Vec3 minp, Vec3 d, RXREAL *dF);

	// SPHo
	virtual void OutputSetting(string fn);

	// `֐
	virtual void DrawCell(int i, int j, int k);
	virtual void DrawCells(Vec3 col, Vec3 col2, int sel = 0);
	virtual void DrawObstacles(void);

protected:
	virtual void setObjectToCell(RXREAL *p);


public:
	// SPH
	void Initialize(const rxSPHEnviroment &env);
	void Allocate(int max_particles);
	void Finalize(void);

	// ZɃp[eBNi[
	virtual void SetParticlesToCell(void);
	virtual void SetPolygonsToCell(void);

	int  GetPolygonsInCell(int gi, int gj, int gk, vector<int> &polys){ return 0; }
	bool IsPolygonsInCell(int gi, int gj, int gk){ return false; }

	void CalMaxDensity(int k);
	
	//
	// Anisotropic kernel
	//
	virtual void CalAnisotropicKernel(void);

protected:
	// Z̏ݒ
	void setupCells(const Vec3 &vMin, const Vec3 &vMax, const double &h);

	// ObhnbV̌vZ
	uint calGridHash(uint x, uint y, uint z);
	uint calGridHash(Vec3 pos);

	// |S𕪊ZɊi[
	void setPolysToCell(RXREAL *vrts, int nv, int* tris, int nt);
};


#endif	// _SPH_H_

